@@ -19,7 +19,6 @@ import (
 	"cmp"
 	"encoding"
 	"encoding/json"
-	"math"
 	"reflect"
 	"slices"
 	"strconv"
@@ -30,7 +29,6 @@ import (
 
 	"cuelang.org/go/cue/errors"
 	"cuelang.org/go/internal/core/adt"
-	"cuelang.org/go/internal/cueexperiment"
 )
 
 // Decode initializes the value pointed to by x with Value v.
@@ -102,7 +100,7 @@ func (d *decoder) decode(x reflect.Value, v Value, isPtr bool) {
 	}
 
 	switch x.Kind() {
-	case reflect.Ptr, reflect.Map, reflect.Slice, reflect.Interface:
+	case reflect.Pointer, reflect.Map, reflect.Slice, reflect.Interface:
 		// nullable types
 		if v.IsNull() || !v.IsConcrete() {
 			d.clear(x)
@@ -146,7 +144,7 @@ func (d *decoder) decode(x reflect.Value, v Value, isPtr bool) {
 	}
 
 	switch kind {
-	case reflect.Ptr:
+	case reflect.Pointer:
 		d.decode(x.Elem(), v, true)
 
 	case reflect.Bool:
@@ -270,17 +268,7 @@ func (d *decoder) interfaceValue(v Value) (x interface{}) {
 
 	case IntKind:
 		if i, err := v.Int64(); err == nil {
-			cueexperiment.Init()
-			if cueexperiment.Flags.DecodeInt64 {
-				return i
-			}
-			// When the decodeint64 experiment is not enabled, we want to return the value
-			// as `int`, but that's not possible for large values on 32-bit architectures.
-			// To avoid overflows causing entirely wrong values to be returned to the user,
-			// let the logic continue below so that we return a *big.Int instead.
-			if i <= math.MaxInt && i >= math.MinInt {
-				return int(i)
-			}
+			return i
 		}
 		x, err = v.Int(nil)
 
@@ -310,7 +298,7 @@ func (d *decoder) interfaceValue(v Value) (x interface{}) {
 		iter, err := v.Fields()
 		d.addErr(err)
 		for iter.Next() {
-			m[iter.Label()] = d.interfaceValue(iter.Value())
+			m[iter.Selector().Unquoted()] = d.interfaceValue(iter.Value())
 		}
 		x = m
 
@@ -352,7 +340,7 @@ func (d *decoder) convertMap(x reflect.Value, v Value) {
 	iter, err := v.Fields()
 	d.addErr(err)
 	for iter.Next() {
-		key := iter.Label()
+		key := iter.Selector().Unquoted()
 
 		var kv reflect.Value
 		kt := t.Key()
@@ -368,7 +356,7 @@ func (d *decoder) convertMap(x reflect.Value, v Value) {
 			case reflect.Int, reflect.Int8, reflect.Int16, reflect.Int32, reflect.Int64:
 				n, err := strconv.ParseInt(key, 10, 64)
 				d.addErr(err)
-				if reflect.Zero(kt).OverflowInt(n) {
+				if kt.OverflowInt(n) {
 					d.addErr(errors.Newf(v.Pos(), "key integer %d overflows %s", n, kt))
 					break
 				}
@@ -377,7 +365,7 @@ func (d *decoder) convertMap(x reflect.Value, v Value) {
 			case reflect.Uint, reflect.Uint8, reflect.Uint16, reflect.Uint32, reflect.Uint64, reflect.Uintptr:
 				n, err := strconv.ParseUint(key, 10, 64)
 				d.addErr(err)
-				if reflect.Zero(kt).OverflowUint(n) {
+				if kt.OverflowUint(n) {
 					d.addErr(errors.Newf(v.Pos(), "key integer %d overflows %s", n, kt))
 					break
 				}
@@ -408,9 +396,8 @@ func (d *decoder) convertStruct(x reflect.Value, v Value) {
 	iter, err := v.Fields()
 	d.addErr(err)
 	for iter.Next() {
-
 		var f *goField
-		key := iter.Label()
+		key := iter.Selector().Unquoted()
 		if i, ok := fields.nameIndex[key]; ok {
 			// Found an exact name match.
 			f = &fields.list[i]
@@ -434,7 +421,7 @@ func (d *decoder) convertStruct(x reflect.Value, v Value) {
 		// Figure out field corresponding to key.
 		subv := x
 		for _, i := range f.index {
-			if subv.Kind() == reflect.Ptr {
+			if subv.Kind() == reflect.Pointer {
 				if subv.IsNil() {
 					// If a struct embeds a pointer to an unexported type,
 					// it is not possible to set a newly allocated value
@@ -538,12 +525,12 @@ func typeFields(t reflect.Type) structFields {
 			visited[f.typ] = true
 
 			// Scan f.typ for fields to include.
-			for i := 0; i < f.typ.NumField(); i++ {
+			for i := range f.typ.NumField() {
 				sf := f.typ.Field(i)
 				isUnexported := sf.PkgPath != ""
 				if sf.Anonymous {
 					t := sf.Type
-					if t.Kind() == reflect.Ptr {
+					if t.Kind() == reflect.Pointer {
 						t = t.Elem()
 					}
 					if isUnexported && t.Kind() != reflect.Struct {
@@ -560,7 +547,7 @@ func typeFields(t reflect.Type) structFields {
 				if tag == "-" {
 					continue
 				}
-				name, opts := parseTag(tag)
+				name, opts, _ := strings.Cut(tag, ",")
 				if !isValidTag(name) {
 					name = ""
 				}
@@ -569,7 +556,7 @@ func typeFields(t reflect.Type) structFields {
 				index[len(f.index)] = i
 
 				ft := sf.Type
-				if ft.Name() == "" && ft.Kind() == reflect.Ptr {
+				if ft.Name() == "" && ft.Kind() == reflect.Pointer {
 					// Follow pointer.
 					ft = ft.Elem()
 				}
@@ -585,7 +572,7 @@ func typeFields(t reflect.Type) structFields {
 						tag:       tagged,
 						index:     index,
 						typ:       ft,
-						omitEmpty: opts.Contains("omitempty"),
+						omitEmpty: tagOptions(opts).Contains("omitempty"),
 					}
 					field.nameBytes = []byte(field.name)
 					field.equalFold = foldFunc(field.nameBytes)
@@ -699,15 +686,6 @@ func cachedTypeFields(t reflect.Type) structFields {
 // tag, or the empty string. It does not include the leading comma.
 type tagOptions string
 
-// parseTag splits a struct field's json tag into its name and
-// comma-separated options.
-func parseTag(tag string) (string, tagOptions) {
-	if idx := strings.Index(tag, ","); idx != -1 {
-		return tag[:idx], tagOptions(tag[idx+1:])
-	}
-	return tag, tagOptions("")
-}
-
 // Contains reports whether a comma-separated list of options
 // contains a particular substr flag. substr must be surrounded by a
 // string boundary or commas.
@@ -718,10 +696,7 @@ func (o tagOptions) Contains(optionName string) bool {
 	s := string(o)
 	for s != "" {
 		var next string
-		i := strings.Index(s, ",")
-		if i >= 0 {
-			s, next = s[:i], s[i+1:]
-		}
+		s, next, _ = strings.Cut(s, ",")
 		if s == optionName {
 			return true
 		}
@@ -884,7 +859,7 @@ func indirect(v reflect.Value, decodingNull bool) (json.Unmarshaler, encoding.Te
 	// If v is a named type and is addressable,
 	// start with its address, so that if the type has pointer methods,
 	// we find them.
-	if v.Kind() != reflect.Ptr && v.Type().Name() != "" && v.CanAddr() {
+	if v.Kind() != reflect.Pointer && v.Type().Name() != "" && v.CanAddr() {
 		haveAddr = true
 		v = v.Addr()
 	}
@@ -893,14 +868,14 @@ func indirect(v reflect.Value, decodingNull bool) (json.Unmarshaler, encoding.Te
 		// usefully addressable.
 		if v.Kind() == reflect.Interface && !v.IsNil() {
 			e := v.Elem()
-			if e.Kind() == reflect.Ptr && !e.IsNil() && (!decodingNull || e.Elem().Kind() == reflect.Ptr) {
+			if e.Kind() == reflect.Pointer && !e.IsNil() && (!decodingNull || e.Elem().Kind() == reflect.Pointer) {
 				haveAddr = false
 				v = e
 				continue
 			}
 		}
 
-		if v.Kind() != reflect.Ptr {
+		if v.Kind() != reflect.Pointer {
 			break
 		}
 
