@@ -57,7 +57,7 @@ func init() {
 	handleComprehension = &runner{
 		name:      "Comprehension",
 		f:         processComprehension,
-		completes: valueKnown | allTasksCompleted | fieldConjunctsKnown,
+		completes: valueKnown | allTasksCompleted | fieldConjunctsKnown | pendingKnown,
 	}
 	handleListLit = &runner{
 		name:      "ListLit",
@@ -101,7 +101,7 @@ func processResolver(ctx *OpContext, t *task, mode runMode) {
 	// be conclusive, we could avoid triggering evaluating disjunctions. This
 	// would be a pretty significant rework, though.
 
-	arc := r.resolve(ctx, oldOnly(0))
+	arc := r.resolve(ctx, combineMode(fieldSetKnown, mode))
 	// TODO: ensure that resolve always returns one of these two.
 	if arc == nil || arc == emptyNode {
 		// TODO: yield instead?
@@ -109,7 +109,9 @@ func processResolver(ctx *OpContext, t *task, mode runMode) {
 	}
 	arc = arc.DerefNonDisjunct()
 
-	ctx.Logf(t.node.node, "RESOLVED %v to %v %v", r, arc.Label, fmt.Sprintf("%p", arc))
+	if ctx.LogEval > 0 {
+		ctx.Logf(t.node.node, "RESOLVED %v to %v %v", r, arc.Label, fmt.Sprintf("%p", arc))
+	}
 	// TODO: consider moving after markCycle or removing.
 	d := arc.DerefDisjunct()
 
@@ -119,6 +121,9 @@ func processResolver(ctx *OpContext, t *task, mode runMode) {
 	// we are done computing and we can return the arc as is.
 	ci, skip := t.node.detectCycleV3(d, t.env, r, ci)
 	if skip {
+		// Either we have a structure cycle or we are unifying with another
+		// conjunct. In either case, we are no longer structure sharing here.
+		t.node.unshare()
 		return
 	}
 
@@ -142,7 +147,7 @@ func processDynamic(ctx *OpContext, t *task, mode runMode) {
 
 	field := t.x.(*DynamicField)
 
-	v := ctx.scalarValue(t, field.Key)
+	v := ctx.value(field.Key, combineMode(scalarValue, mode))
 	if v == nil {
 		return
 	}
@@ -169,8 +174,6 @@ func processDynamic(ctx *OpContext, t *task, mode runMode) {
 	ci := t.id
 
 	c := MakeConjunct(t.env, field, ci)
-	// TODO(evalv3): this does not seem to be necessary and even treacherous.
-	c.CloseInfo.cc = nil
 	n.insertArc(f, field.ArcType, c, ci, true)
 }
 
@@ -181,7 +184,7 @@ func processPatternConstraint(ctx *OpContext, t *task, mode runMode) {
 
 	// Note that the result may be a disjunction. Be sure to not take the
 	// default value as we want to retain the options of the disjunction.
-	v := ctx.evalState(field.Filter, require(0, scalarKnown))
+	v := ctx.evalState(field.Filter, require(0, scalarValue))
 	if v == nil {
 		return
 	}
@@ -215,11 +218,6 @@ func processDisjunctions(c *OpContext, t *task, mode runMode) {
 	t.err = CombineErrors(nil, t.err, err)
 }
 
-func processFinalizeDisjunctions(c *OpContext, t *task, mode runMode) {
-	n := t.node
-	n.finalizeDisjunctions()
-}
-
 func processListLit(c *OpContext, t *task, mode runMode) {
 	n := t.node
 
@@ -252,6 +250,9 @@ func processListLit(c *OpContext, t *task, mode runMode) {
 			}
 
 		case *Ellipsis:
+			// TODO(openlist): this will work once we have the same closedness
+			// semantics for lists as for structs.
+			// t.id.cc.isTotal = true
 			if j != len(l.Elems)-1 {
 				n.addErr(c.Newf("ellipsis must be last element in list"))
 				return
@@ -344,7 +345,6 @@ func processListVertex(c *OpContext, t *task, mode runMode) {
 			continue
 		}
 		for _, c := range a.Conjuncts {
-			c.CloseInfo.cc = t.id.cc
 			n.insertArc(a.Label, ArcMember, c, t.id, true)
 		}
 	}
