@@ -20,9 +20,11 @@ import (
 	"fmt"
 	"io"
 	"io/fs"
+	"maps"
 	"os"
 	"path"
 	"path/filepath"
+	"slices"
 	"strings"
 	"testing"
 
@@ -82,7 +84,7 @@ type TxTarTest struct {
 
 // A Test represents a single test based on a .txtar file.
 //
-// A Test embeds *[testing.T] and should be used to report errors.
+// A Test embeds [*testing.T] and should be used to report errors.
 //
 // Entries within the txtar file define CUE files (available via the
 // Instances and RawInstances methods) and expected output
@@ -96,6 +98,7 @@ type TxTarTest struct {
 // A txtar file can define test-specific tags and values in the comment section.
 // These are available via the [Test.HasTag] and [Test.Value] methods.
 // The #skip tag causes a [Test] to be skipped.
+// When running via [cuetdtest.Matrix], #skip-[cuetdtest.M.Name] tags can also be used.
 // The #noformat tag causes the $CUE_FORMAT_TXTAR value
 // to be ignored.
 //
@@ -125,6 +128,10 @@ type Test struct {
 	hasGold bool
 }
 
+// Ensure that Test always implements testing.TB.
+// Note that testing.TB may gain new methods in future Go releases.
+var _ testing.TB = (*Test)(nil)
+
 // Write implements [io.Writer] by writing to the output for the test,
 // which will be tested against the main golden file.
 func (t *Test) Write(b []byte) (n int, err error) {
@@ -142,6 +149,20 @@ type file struct {
 	diff     bool // true if this contains a diff between fallback and main
 }
 
+// bytes returns the bytes in the file's buffer, and ensures that the
+// slice finishes with a newline (\n). txtar archives cannot contain
+// files without a final newline. Consequently, when comparing
+// proposed/generated file content with content from an archive's
+// file, we must ensure that the proposed content also finishes with a
+// newline.
+func (f *file) bytes() []byte {
+	bs := f.buf.Bytes()
+	if l := len(bs); l > 0 && bs[l-1] != '\n' {
+		bs = append(bs, '\n')
+	}
+	return bs
+}
+
 // HasTag reports whether the tag with the given key is defined
 // for the current test. A tag x is defined by a line in the comment
 // section of the txtar file like:
@@ -320,10 +341,10 @@ func loadWithConfig(a *txtar.Archive, dir string, cfg load.Config, args ...strin
 // for more details.
 func (x *TxTarTest) Run(t *testing.T, f func(tc *Test)) {
 	if x.Matrix == nil {
-		x.run(t, f)
+		x.run(t, nil, f)
 		return
 	}
-	x.Matrix.Do(t, func(m *cuetdtest.M) {
+	x.Matrix.Do(t, func(t *testing.T, m *cuetdtest.M) {
 		test := *x
 		if s := m.Fallback(); s != "" {
 			test.Fallback = test.Name
@@ -334,8 +355,7 @@ func (x *TxTarTest) Run(t *testing.T, f func(tc *Test)) {
 		if s := m.Name(); s != cuetdtest.DefaultVersion {
 			test.Name += "-" + s
 		}
-		test.run(m.T, func(tc *Test) {
-			tc.M = m
+		test.run(t, m, func(tc *Test) {
 			f(tc)
 		})
 	})
@@ -343,18 +363,18 @@ func (x *TxTarTest) Run(t *testing.T, f func(tc *Test)) {
 
 // Runtime returns a new runtime based on the configuration of the test.
 func (t *Test) Runtime() *runtime.Runtime {
-	return (*runtime.Runtime)(t.Context())
+	return (*runtime.Runtime)(t.CueContext())
 }
 
-// Context returns a new cue.Context based on the configuration of the test.
-func (t *Test) Context() *cue.Context {
+// CueContext returns a new cue.CueContext based on the configuration of the test.
+func (t *Test) CueContext() *cue.Context {
 	if t.M != nil {
-		return t.M.Context()
+		return t.M.CueContext()
 	}
 	return cuecontext.New()
 }
 
-func (x *TxTarTest) run(t *testing.T, f func(tc *Test)) {
+func (x *TxTarTest) run(t *testing.T, m *cuetdtest.M, f func(tc *Test)) {
 	t.Helper()
 
 	if x.DebugArchive != "" {
@@ -366,6 +386,7 @@ func (x *TxTarTest) run(t *testing.T, f func(tc *Test)) {
 			}
 			tc := &Test{
 				T:       t,
+				M:       m,
 				Archive: archive,
 				Dir:     "/tmp",
 
@@ -416,8 +437,10 @@ func (x *TxTarTest) run(t *testing.T, f func(tc *Test)) {
 			if err != nil {
 				t.Fatalf("error parsing txtar file: %v", err)
 			}
+
 			tc := &Test{
 				T:       t,
+				M:       m,
 				Archive: a,
 				Dir:     filepath.Dir(filepath.Join(dir, fullpath)),
 
@@ -437,7 +460,16 @@ func (x *TxTarTest) run(t *testing.T, f func(tc *Test)) {
 			if tc.HasTag("skip") {
 				t.Skip()
 			}
-
+			if tc.M != nil {
+				// When running via [cuetdtest.Matrix], support e.g. #skip-v2.
+				if tc.HasTag("skip-" + tc.Name()) {
+					t.Skip()
+				}
+			} else if tc.HasTag("skip-v2") && strings.Contains(t.Name(), "EvalV2") {
+				// Temporary hack since internal/core/adt uses TestEvalV2 rather than [cuetdtest.Matrix].
+				// TODO(mvdan): clean this up.
+				t.Skip()
+			}
 			if msg, ok := x.Skip[testName]; ok {
 				t.Skip(msg)
 			}
@@ -472,11 +504,19 @@ func (x *TxTarTest) run(t *testing.T, f func(tc *Test)) {
 			}
 			f(tc)
 
+			// Track the position of the fallback files.
 			index := make(map[string]int, len(a.Files))
 			for i, f := range a.Files {
+				if _, ok := index[f.Name]; ok {
+					t.Errorf("duplicated txtar file entry %s", f.Name)
+				}
 				index[f.Name] = i
 			}
 
+			// Record ordering of files in the archive to preserve that ordering
+			// later.
+			ordering := maps.Clone(index)
+
 			// Add diff files between fallback and main file. These are added
 			// as regular output files so that they can be updated as well.
 			for _, sub := range tc.outFiles {
@@ -484,14 +524,20 @@ func (x *TxTarTest) run(t *testing.T, f func(tc *Test)) {
 					continue
 				}
 				if j, ok := index[sub.fallback]; ok {
+					if _, ok := ordering[sub.name]; !ok {
+						ordering[sub.name] = j
+					}
 					fallback := a.Files[j].Data
 
-					result := sub.buf.Bytes()
+					result := sub.bytes()
 					if len(result) == 0 || len(fallback) == 0 {
 						continue
 					}
 
 					diffName := "diff/-" + sub.name + "<==>+" + sub.fallback
+					if _, ok := ordering[diffName]; !ok {
+						ordering[diffName] = j
+					}
 					switch diff := diff.Diff("old", fallback, "new", result); {
 					case len(diff) > 0:
 						tc.outFiles = append(tc.outFiles, file{
@@ -530,30 +576,10 @@ func (x *TxTarTest) run(t *testing.T, f func(tc *Test)) {
 				}
 			}
 
-			// Insert results of this test at first location of any existing
-			// test or at end of list otherwise.
-			k := len(a.Files)
-			for _, sub := range tc.outFiles {
-				if i, ok := index[sub.name]; ok {
-					k = i
-					break
-				}
-				if i, ok := index[sub.fallback]; ok {
-					k = i
-					break
-				}
-			}
-
-			// Filter files up to k from the original list.
 			files := make([]txtar.File, 0, len(a.Files))
-			for _, f := range a.Files[:k] {
-				if _, ok := index[f.Name]; ok {
-					files = append(files, f)
-				}
-			}
 
 			for _, sub := range tc.outFiles {
-				result := sub.buf.Bytes()
+				result := sub.bytes()
 
 				files = append(files, txtar.File{Name: sub.name})
 				gold := &files[len(files)-1]
@@ -597,7 +623,7 @@ func (x *TxTarTest) run(t *testing.T, f func(tc *Test)) {
 
 			// Add remaining unrelated files, ignoring files that were already
 			// added.
-			for _, f := range a.Files[k:] {
+			for _, f := range a.Files {
 				if _, ok := index[f.Name]; ok {
 					files = append(files, f)
 				}
@@ -605,6 +631,18 @@ func (x *TxTarTest) run(t *testing.T, f func(tc *Test)) {
 			a.Files = files
 
 			if update {
+				slices.SortStableFunc(a.Files, func(i, j txtar.File) int {
+					p, ok := ordering[i.Name]
+					if !ok {
+						p = len(a.Files)
+					}
+					q, ok := ordering[j.Name]
+					if !ok {
+						q = len(a.Files)
+					}
+					return p - q
+				})
+
 				err = os.WriteFile(fullpath, txtar.Format(a), 0644)
 				if err != nil {
 					t.Fatal(err)
